/*
 *  DataMiningTask.java
 *  Copyright (C) 2007 University of Calabria
 *  Dept. of Electronics, Computer Science and Systems
 */

package weka4ws.service.impl;

import java.io.BufferedInputStream;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.Random;
import java.util.StringTokenizer;
import java.util.zip.CRC32;
import java.util.zip.Checksum;

import org.apache.log4j.Level;
import org.globus.security.gridmap.GridMap;
import org.globus.wsrf.security.SecurityManager;

import weka4ws.stubs.service.Algorithm;
import weka4ws.stubs.service.Dataset;
import weka4ws.stubs.service.Result;
import weka4ws.stubs.service.Weka4WSException;

/**
 * @author <a href="mailto:mlackovic@deis.unical.it">Marco Lackovic</a>
 */
public abstract class DataMiningTask implements DataMiningConstants {

    /* Web service fields */

    protected Algorithm algorithm;

    protected Dataset dataset;

    protected String userID;

    protected long taskID;

    protected ServiceParameters parameters;

    protected Dataset tempDataset = new Dataset();

    protected Weka4WSResource resultResource;

    protected Result result = new Result();

    protected Weka4WSException wekaException = new Weka4WSException();

    /* Data mining fields */

    /** A thread that classification runs in */
    protected Thread m_RunThread;

    public DataMiningTask(ServiceParameters parameters, String type,
            Weka4WSResource resultResource) throws Exception {

        if (WUtil.logging) {
            logger.setLevel(Level.ALL);
        } else {
            logger.setLevel(Level.OFF);
        }
        taskID = parameters.getTaskID();
        algorithm = parameters.getAlgorithm();
        dataset = parameters.getDataset();
        this.parameters = parameters;

        resourceInitialization(resultResource);

        String identity = SecurityManager.getManager().getCaller();
        GridMap map = new GridMap();
        map.load(GRIDMAPFILE);
        userID = map.getUserID(identity);

        dataset.setDirPath(userHomeDir(userID) + File.separator + TEMP_DIR
                + File.separator);

        dataset.setFilePath(dataset.getDirPath() + dataset.getFileName());

        logger.info("'" + identity + "' mapped to '" + userID + "'");

        logger.info("'" + userID + "' invoked " + type + " on '"
                + dataset.getFileName() + "'");

        /*        try {
         SecurityManager.getManager().setServiceOwnerFromContext();
         } catch (Exception e) {
         logger
         .warn("Unable to set service owner from context (check that credentials have been delegated)");
         }*/

        createUserStorageDir();
    }

    private void resourceInitialization(Weka4WSResource resultResource) {
        result.setReady(false);
        wekaException.setThrown(false);
        this.resultResource = resultResource;
        this.resultResource.setResult(result);
    }

    /**
     * Gets the local home directory of the remote caller by reading the "/etc/passwd" file.
     */
    private String userHomeDir(String caller) throws Exception {
        Runtime runtime = Runtime.getRuntime();
        Process proc = runtime.exec("grep " + caller + " /etc/passwd"); // starts the grep command
        InputStream inputStream = proc.getInputStream();
        InputStreamReader inputStreamReader = new InputStreamReader(inputStream);
        BufferedReader bufferedReader = new BufferedReader(inputStreamReader);
        // reads the grep output
        String line;
        String lineHome = "";
        while ((line = bufferedReader.readLine()) != null) {
            lineHome = line;
        }
        proc.waitFor();

        if (lineHome.length() == 0) {
            Exception ex = new Exception("User " + caller
                    + " doesn't exist on this machine");
            throw ex;
        }

        lineHome = lineHome.substring(0, lineHome.lastIndexOf(':'));
        lineHome = lineHome.substring(lineHome.lastIndexOf(':') + 1, lineHome
                .length());
        logger.debug("lineHome = " + lineHome);

        return lineHome;
    }

    /**
     * Creates the storage directory in the local home directory of the caller.
     */
    private void createUserStorageDir() {
        try {
            Runtime rt = Runtime.getRuntime();
            String command = "sudo mkdir " + dataset.getDirPath();
            Process proc1 = rt.exec(command);
            proc1.waitFor();
            command = "sudo chown -R " + userID + ". " + dataset.getDirPath();
            Process proc2 = rt.exec(command);
            proc2.waitFor();
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    /**
     * Removes the dataset from the temporary storage directory of the "globus" user.
     */
    protected static void removeFromTempDir(String filePath) {
        try {
            Runtime rt = Runtime.getRuntime();
            Process proc = rt.exec("rm -f " + filePath); // deletes the file
            proc.waitFor();
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    protected void finalOperations(Result result, Weka4WSException wekaException) {
        result.setException(wekaException);
        result.setReady(true);
        resultResource.setResult(result);
        String tempDatasetFilePath = tempDataset.getFilePath();
        if ((tempDatasetFilePath != null) && (tempDatasetFilePath != "")) {
            removeFromTempDir(tempDatasetFilePath);
        }
        Weka4WSService.removeTask(taskID);
    }

    /**
     * Checks if a file is available in a directory.
     * @throws IOException
     */
    protected boolean isAvailable(Dataset dataset, Dataset tempDataset)
            throws Exception {

        Runtime runtime = Runtime.getRuntime();
        String command = "sudo ls " + dataset.getDirPath();
        Process proc = runtime.exec(command);
        InputStream inputStream = proc.getInputStream();
        InputStreamReader inputStreamReader = new InputStreamReader(inputStream);
        BufferedReader bufferedReader = new BufferedReader(inputStreamReader);
        String line;
        String fileName = dataset.getFileName();

        while ((line = bufferedReader.readLine()) != null) {
            if (line.startsWith(fileName)) {
                break;
            }
        }

        if (line != null) {
            String filePath = dataset.getFilePath();
            if (line.endsWith(".gz")) {
                filePath += ".gz";
                if (isFileEmpty(filePath)) {
                    String errorMessage = "File '" + filePath
                            + "' is empty; please remove it manually.\n";
                    throw new Exception(errorMessage);
                }
                logger.info("Compressed file '" + fileName
                        + ".gz' found. Decompressing...");
                extract(dataset);
                logger.info("File '" + fileName + ".gz' decompressed.");

            } else if (isFileEmpty(filePath)) {
                String errorMessage = "File '" + filePath
                        + "' is empty; please remove it manually.\n";
                throw new Exception(errorMessage);
            }

            String extension = fileName.substring(fileName.lastIndexOf("."));

            tempDataset.setFileName(Math.abs((new Random()).nextLong())
                    + extension);

            String globusHomeDir = System.getProperties().getProperty(
                    "user.home");
            tempDataset.setDirPath(globusHomeDir + File.separator + TEMP_DIR
                    + File.separator);

            tempDataset.setFilePath(tempDataset.getDirPath()
                    + tempDataset.getFileName());

            File tempDirectory = new File(tempDataset.getDirPath());

            if (!tempDirectory.exists()) {
                tempDirectory.mkdir();
            }

            copyToTempDir(dataset.getFilePath(), tempDataset.getFilePath());

            long cs = getChecksumValue(tempDataset.getFilePath());
            if (cs != dataset.getCrc()) {
                logger.info("Local file '" + dataset.getFileName()
                        + "' CRC check failed: transfer needed.\n");
                removeFromTempDir(tempDataset.getFileName());
                return false;
            }
            logger.info("Local file '" + dataset.getFileName()
                    + "' CRC check passed: transfer not needed.");

            return true;
        }
        logger.info("Local file '" + dataset.getFileName()
                + "' not found: transfer needed.\n");

        return false;
    }

    private boolean isFileEmpty(String filePath) throws IOException {

        Runtime runtime = Runtime.getRuntime();
        String command = "sudo ls -l " + filePath;
        Process proc = runtime.exec(command);
        InputStream inputStream = proc.getInputStream();
        InputStreamReader inputStreamReader = new InputStreamReader(inputStream);
        BufferedReader bufferedReader = new BufferedReader(inputStreamReader);
        String line;

        while ((line = bufferedReader.readLine()) != null) {
            int i = 0;
            String sizeString = "";
            StringTokenizer st = new StringTokenizer(line);
            while (i < 5 && st.hasMoreTokens()) {
                i++;
                sizeString = st.nextToken();
            }
            if (Integer.parseInt(sizeString) == 0) {
                return true;
            }
        }
        return false;
    }

    public boolean isDatasetAvailable() throws Exception {
        return isAvailable(dataset, tempDataset);
    }

    protected void extract(Dataset dataset) {

        try {
            Runtime rt = Runtime.getRuntime();
            String command = "sudo gzip -d " + dataset.getFilePath() + ".gz";
            logger.debug(command);
            Process proc1 = rt.exec(command);
            proc1.waitFor();
            command = "sudo chown " + userID + ". " + dataset.getFilePath();
            logger.debug(command);
            Process proc2 = rt.exec(command);
            proc2.waitFor();
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    /**
     * Copies the dataset from the user storage directory 
     * to the temporary storage directory of the "globus" user. 
     */
    private void copyToTempDir(String sourceFilePath, String destFilePath) {
        try {
            Runtime rt = Runtime.getRuntime();
            Process proc1 = rt.exec("sudo cp " + sourceFilePath + " "
                    + destFilePath); // copies the file
            proc1.waitFor();
            String globusUsername = System.getProperties().getProperty(
                    "user.name");
            Process proc2 = rt.exec("sudo chown " + globusUsername + ". "
                    + destFilePath); // changes the owner to "globus"
            proc2.waitFor();
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    /**
     * Given a file path returns the file CRC.
     */
    private long getChecksumValue(String filePath) {
        Checksum checksum = new CRC32();
        try {
            BufferedInputStream is = new BufferedInputStream(
                    new FileInputStream(filePath));
            byte[] bytes = new byte[1024];
            int len = 0;
            while ((len = is.read(bytes)) >= 0) {
                checksum.update(bytes, 0, len);
            }
            is.close();
        } catch (IOException e) {
            e.printStackTrace();
        }
        return checksum.getValue();
    }

    public String getSavePath() {
        return dataset.getDirPath();
    }

    /**
     * Stops the running task corresponding to the passed ID.
     */
    protected String stopTask() {

        if (m_RunThread != null) {
            m_RunThread.interrupt();
            // This is deprecated (and theoretically the interrupt should do).
            m_RunThread.stop();

            removeFromTempDir(tempDataset.getFilePath());

            logger.info("Task stopped");
            return "Task stopped";
        }
        logger.info("Task thread has already terminated");
        return "Task thread has already terminated";
    }
}
